from typing import Union, Dict, List, Tuple
from copy import deepcopy
from math import acos, ceil
import numpy as np
import networkx as nx

from . import NML
from .nml_generation import generate_graph, generate_nml


def get_vector(node: Dict, scale) -> np.ndarray:
    return np.array(node["position"]) * scale


def vector_length(vector: np.ndarray) -> np.ndarray:
    return np.sqrt(vector.dot(vector))


def get_vector_between_nodes(node1: Dict, node2: Dict, scale) -> np.ndarray:
    return get_vector(node2, scale) - get_vector(node1, scale)


def calculate_angle_between_vectors(vector1: np.ndarray, vector2: np.ndarray) -> float:
    dot_val = vector1.dot(vector2)
    length_values = vector_length(vector1) * vector_length(vector2)
    angle = acos(dot_val / length_values)
    return angle


def calculate_distance_between_nodes(
    node1: Dict, node2: Dict, scale: np.ndarray
) -> float:
    difference_vector = get_vector_between_nodes(node1, node2, scale)
    return vector_length(difference_vector)


# First calculate the vector from node1 to node2 and then scale it with the percentage of scale.
# Therefore it returns a position along the vector between node1 and node2.
# The Position along the vector is determined by the scale.
# It is used to calculate the padding node position which is needed to ensure the maximum length.
def get_padding_node_position(
    node1: Dict, node2: Dict, relative_distance_along_vector: float, scale: np.ndarray
) -> List[int]:
    node1_position = get_vector(node1, scale)
    vector_between_nodes = get_vector_between_nodes(node1, node2, scale)
    return (
        node1_position + vector_between_nodes * relative_distance_along_vector
    ) / scale


def detect_max_node_id_from_all_graphs(graph_dict: Dict[str, nx.Graph]) -> int:
    max_id = 0
    for group in graph_dict.values():
        for tree in group:
            max_id_of_current_tree = np.array(list(tree.nodes)).max()
            max_id = max(max_id_of_current_tree, max_id)
    return max_id


def approximate_minimal_edge_length(
    nml_or_graph: Union[NML, Tuple[Dict[str, List[nx.Graph]], Dict]],
    max_length: float,
    max_angle: float,
) -> Union[NML, Tuple[Dict[str, List[nx.Graph]], Dict]]:
    # it is easier to operate on a graph
    if isinstance(nml_or_graph, NML):
        nml_graph, parameter_dict = generate_graph(nml_or_graph)
    else:
        nml_graph, parameter_dict = nml_or_graph
    scale = np.array(parameter_dict["scale"])
    for group in nml_graph.values():
        for graph in group:
            approximate_minimal_edge_length_for_graph(
                graph, max_length, max_angle, scale
            )

    # return the same format as the input
    if isinstance(nml_or_graph, NML):
        return generate_nml(nml_graph, parameter_dict, globalize_ids=False)
    else:
        return nml_graph, parameter_dict


def approximate_minimal_edge_length_for_graph(
    graph: nx.Graph, max_length: float, max_angle: float, scale: np.ndarray
):
    nodes_with_degree_of_two = [node for node in graph.nodes if graph.degree(node) == 2]

    for node_id in nodes_with_degree_of_two:
        current_node = graph.nodes[node_id]
        neighbors = list(graph.neighbors(node_id))
        neighbor1 = graph.nodes[neighbors[0]]
        neighbor2 = graph.nodes[neighbors[1]]
        vector1 = get_vector_between_nodes(neighbor1, current_node, scale)
        vector2 = get_vector_between_nodes(current_node, neighbor2, scale)
        new_edge_vector = get_vector_between_nodes(neighbor1, neighbor2, scale)
        distance_between_combined_edges = vector_length(new_edge_vector)
        angle = calculate_angle_between_vectors(vector1, vector2)
        if angle <= max_angle and distance_between_combined_edges <= max_length:
            graph.remove_edges_from([(neighbors[0], node_id), (node_id, neighbors[1])])
            graph.remove_node(node_id)
            graph.add_edge(neighbors[0], neighbors[1])


def ensure_max_edge_length(
    nml_or_graph: Union[NML, Tuple[Dict[str, List[nx.Graph]], Dict]], max_length: float
) -> Union[NML, Tuple[Dict[str, List[nx.Graph]], Dict]]:
    # it is easier to operate on a graph
    if isinstance(nml_or_graph, NML):
        nml_graph, parameter_dict = generate_graph(nml_or_graph)
    else:
        nml_graph, parameter_dict = nml_or_graph
    scale = np.array(parameter_dict["scale"])
    max_id = detect_max_node_id_from_all_graphs(nml_graph)
    next_valid_id = max_id + 1
    for group in nml_graph.values():
        for graph in group:
            next_valid_id = ensure_max_edge_length_for_graph(
                graph, max_length, next_valid_id, scale
            )

    # return the same format as the input
    if isinstance(nml_or_graph, NML):
        return generate_nml(nml_graph, parameter_dict, globalize_ids=False)
    else:
        return nml_graph, parameter_dict


def ensure_max_edge_length_for_graph(
    graph: nx.Graph, max_length: float, current_id: int, scale: np.ndarray
) -> int:
    edges_to_be_added = []
    edges_to_be_removed = []
    nodes_to_be_added = []
    for edge in graph.edges:
        node1 = graph.nodes[edge[0]]
        node2 = graph.nodes[edge[1]]
        edge_distance = calculate_distance_between_nodes(node1, node2, scale)
        # add padding nodes if the distance is too high
        if edge_distance > max_length:
            number_of_nodes = ceil(edge_distance / max_length)
            # remove old edge
            edges_to_be_removed.append((edge[0], edge[1]))
            # add all padding nodes and the edges
            previous_node_id = edge[0]
            for padding_node_number in range(1, number_of_nodes):
                relative_distance_between_nodes = padding_node_number / number_of_nodes
                padding_node_position = get_padding_node_position(
                    node1, node2, relative_distance_between_nodes, scale
                )

                # attributes of the new node
                padding_node_attributes = deepcopy(node1)
                padding_node_attributes["position"] = (
                    padding_node_position[0],
                    padding_node_position[1],
                    padding_node_position[2],
                )
                padding_node_attributes["id"] = current_id
                # add node and edge to predecessor
                nodes_to_be_added.append(padding_node_attributes)
                edges_to_be_added.append((previous_node_id, current_id))

                # update variables
                previous_node_id = current_id
                current_id += 1
            # add the edge between the last padding node and the second original node
            edges_to_be_added.append((previous_node_id, edge[1]))

    graph.remove_edges_from(edges_to_be_removed)
    graph.add_edges_from(edges_to_be_added)
    for node in nodes_to_be_added:
        graph.add_node(node["id"], **node)

    return current_id



from wknml import NMLParameters, Group, Edge, Node, Tree, NML, Branchpoint, Comment
import networkx as nx
import numpy as np

import logging
import colorsys
from typing import Tuple, List, Dict, Union, Any
from copy import deepcopy


logger = logging.getLogger(__name__)


def random_color_rgba():
    # https://stackoverflow.com/a/43437435/783758

    h, s, l = (
        np.random.random(),
        0.5 + np.random.random() / 2.0,
        0.4 + np.random.random() / 5.0,
    )
    r, g, b = colorsys.hls_to_rgb(h, l, s)
    return (r, g, b, 1)


def discard_children_hierarchy(groups: List[Group]) -> List[Group]:
    groups_without_hierarchy = []
    for group in groups:
        children = discard_children_hierarchy(group.children)
        groups_without_hierarchy.append(
            Group(id=group.id, name=group.name, children=[])
        )
        groups_without_hierarchy.extend(children)
    return groups_without_hierarchy


def globalize_tree_ids(group_dict: Dict[str, List[nx.Graph]]):
    current_id = 1
    for tree_group in group_dict.values():
        for tree in tree_group:
            tree.graph["id"] = current_id
            current_id += 1


def globalize_node_ids(group_dict: Dict[str, List[nx.Graph]]):
    current_id = 1
    for tree_group in group_dict.values():
        for tree_index in range(len(tree_group)):
            tree = tree_group[tree_index]
            new_tree = nx.Graph(**tree.graph)
            edge_mapping_dict = {}
            for old_id in tree.nodes:
                tree.nodes[old_id]["id"] = current_id
                edge_mapping_dict[old_id] = current_id
                new_tree.add_node(current_id, **tree.nodes[old_id])

                current_id += 1
            new_edges = []
            for edge in tree.edges:
                new_edges.append(
                    (edge_mapping_dict[edge[0]], edge_mapping_dict[edge[1]])
                )

            new_tree.add_edges_from(new_edges)
            tree_group[tree_index] = new_tree


def generate_nml(
    tree_dict: Union[List[nx.Graph], Dict[str, List[nx.Graph]]],
    parameters: Dict[str, Any] = {},
    globalize_ids=True,
) -> NML:
    no_group_provided = False
    if not isinstance(tree_dict, dict):
        tree_dict = {"main_group": tree_dict}
        no_group_provided = True

    if globalize_ids:
        globalize_tree_ids(tree_dict)
        globalize_node_ids(tree_dict)

    nmlParameters = NMLParameters(
        name=parameters.get("name", "dataset"),
        scale=parameters.get("scale", [11.24, 11.24, 25]),
        offset=parameters.get("offset", None),
        time=parameters.get("time", None),
        editPosition=parameters.get("editPosition", None),
        editRotation=parameters.get("editRotation", None),
        zoomLevel=parameters.get("zoomLevel", None),
        taskBoundingBox=parameters.get("taskBoundingBox", None),
        userBoundingBox=parameters.get("userBoundingBox", None),
    )

    comments = [
        Comment(node, tree.nodes[node]["comment"])
        for group in tree_dict.values()
        for tree in group
        for node in tree.nodes
        if "comment" in tree.nodes[node]
    ]

    branchpoints = [
        Branchpoint(tree.nodes[node]["id"], 0)
        for group in tree_dict.values()
        for tree in group
        for node in tree.nodes
        if "branchpoint" in tree.nodes[node]
    ]

    if no_group_provided:
        groups = []
    else:
        groups = [
            Group(id=group_id, name=group_name, children=[])
            for group_id, group_name in enumerate(tree_dict, 1)
        ]

    trees = []

    for group_id, group_name in enumerate(tree_dict, 1):
        for tree in tree_dict[group_name]:
            nodes, edges = extract_nodes_and_edges_from_graph(tree)
            color = tree.graph.get("color", random_color_rgba())
            name = tree.graph.get("name", f"tree{tree.graph['id']}")

            trees.append(
                Tree(
                    nodes=nodes,
                    edges=edges,
                    id=tree.graph["id"],
                    name=name,
                    groupId=None if no_group_provided else group_id,
                    color=color,
                )
            )

    nml = NML(
        parameters=nmlParameters,
        trees=trees,
        branchpoints=branchpoints,
        comments=comments,
        groups=groups,
    )

    return nml


def generate_graph(nml: NML) -> Tuple[Dict[str, List[nx.Graph]], Dict]:
    nml = deepcopy(nml)._replace(groups=discard_children_hierarchy(nml.groups))
    group_dict = {}
    for group in nml.groups:
        graphs_in_current_group = []
        for tree in nml.trees:
            if tree.groupId == group.id:
                graphs_in_current_group.append(nml_tree_to_graph(tree))
        group_dict[group.name] = graphs_in_current_group

    nml_parameters = nml.parameters
    parameter_dict = {}

    parameter_list = [
        "name",
        "scale",
        "offset",
        "time",
        "editPosition",
        "editRotation",
        "zoomLevel",
        "taskBoundingBox",
        "userBoundingBox",
    ]

    for parameter in parameter_list:
        if getattr(nml_parameters, parameter) is not None:
            parameter_dict[parameter] = getattr(nml_parameters, parameter)

    for comment in nml.comments:
        for group in group_dict.values():
            for tree in group:
                if comment.node in tree.nodes:
                    tree.nodes[comment.node]["comment"] = comment.content

    for branchpoint in nml.branchpoints:
        for group in group_dict.values():
            for tree in group:
                if branchpoint.id in tree.nodes:
                    tree.nodes[branchpoint.id]["branchpoint"] = branchpoint.time

    return group_dict, parameter_dict


def nml_tree_to_graph(tree: Tree) -> nx.Graph:
    optional_attribute_list = ["rotation", "inVp", "inMag", "bitDepth", "interpolation"]

    graph = nx.Graph(id=tree.id, color=tree.color, name=tree.name, groupId=tree.groupId)
    for node in tree.nodes:
        node_id = node.id
        graph.add_node(node_id, id=node_id, radius=node.radius, position=node.position)
        for optional_attribute in optional_attribute_list:
            if getattr(node, optional_attribute) is not None:
                graph.nodes[node_id][optional_attribute] = getattr(
                    node, optional_attribute
                )

    graph.add_edges_from([(edge.source, edge.target) for edge in tree.edges])

    return graph


def extract_nodes_and_edges_from_graph(
    graph: nx.Graph,
) -> Tuple[List[Node], List[Edge]]:
    node_nml = [
        Node(
            id=graph.nodes[node]["id"],
            position=graph.nodes[node]["position"],
            radius=graph.nodes[node].get("radius", 1.0),
            rotation=graph.nodes[node].get("rotation", None),
            inVp=graph.nodes[node].get("inVp", None),
            inMag=graph.nodes[node].get("inMag", None),
            bitDepth=graph.nodes[node].get("bitDepth", None),
            interpolation=graph.nodes[node].get("interpolation", None),
            time=graph.nodes[node].get("time", None),
        )
        for node in graph.nodes
    ]

    edge_nml = [Edge(source=edge[0], target=edge[1]) for edge in graph.edges]

    return node_nml, edge_nml



import xml.etree.ElementTree as ET
from loxun import XmlWriter
from typing import BinaryIO, NamedTuple, List, Tuple, Optional

Vector3 = Tuple[float, float, float]
Vector4 = Tuple[float, float, float, float]
IntVector6 = Tuple[int, int, int, int, int, int]


class NMLParameters(NamedTuple):
    """
    Contains common metadata for NML files

    Attributes:
        name: str
        scale: Vector3
        offset: Optional[Vector3]
        time: Optional[int]
        editPosition: Optional[Vector3]
        editRotation: Optional[Vector3]
        zoomLevel: Optional[float]
        taskBoundingBox: Optional[IntVector6]
        userBoundingBox: Optional[IntVector6]
    """

    name: str
    scale: Vector3
    offset: Optional[Vector3]
    time: Optional[int]
    editPosition: Optional[Vector3]
    editRotation: Optional[Vector3]
    zoomLevel: Optional[float]
    taskBoundingBox: Optional[IntVector6]
    userBoundingBox: Optional[IntVector6]


class Node(NamedTuple):
    """
    A webKnossos skeleton node annotation object.

    Attributes:
        id: int
        position: Vector3
        radius: Optional[float]
        rotation: Optional[Vector3]
        inVp: Optional[int]
        inMag: Optional[int]
        bitDepth: Optional[int]
        interpolation: Optional[bool]
        time: Optional[int]
    """

    id: int
    position: Vector3
    radius: Optional[float]
    rotation: Optional[Vector3]
    inVp: Optional[int]
    inMag: Optional[int]
    bitDepth: Optional[int]
    interpolation: Optional[bool]
    time: Optional[int]


class Edge(NamedTuple):
    """
    A webKnossos skeleton edge.

    Attributes:
        source (int): node id reference
        target (int): node id reference
    """

    source: int
    target: int


class Tree(NamedTuple):
    """
    A webKnossos skeleton (tree) object. A graph structure consisting of nodes and edges.

    Attributes:
        id: int
        color (Vector4): RGBA
        name: str
        nodes: List[Node]
        edges: List[Edge]
        groupId (Optional[int]): group id reference
    """

    id: int
    color: Vector4
    name: str
    nodes: List[Node]
    edges: List[Edge]
    groupId: Optional[int]


class Branchpoint(NamedTuple):
    """
    A webKnossos branchpoint, i.e. a skeleton node with more than one outgoing edge.

    Attributes:
        id (int): node id reference
        time (int): Unix timestamp
    """

    id: int
    time: int


class Group(NamedTuple):
    """
    A container to group several skeletons (trees) together. Mostly for cosmetic or organizational purposes.

    Attributes:
        id: int
        name: str
        children: List[Group]
    """

    id: int
    name: str
    children: List["Group"]


class Comment(NamedTuple):
    """
    A single comment belonging to a skeleton node.

    Attributes:
        node (int): node id reference
        content (str): supports Markdown
    """

    node: int
    content: str


class NML(NamedTuple):
    """
    A complete webKnossos skeleton annotation object contain one or more skeletons (trees).

    Attributes:
        parameters: NMLParameters
        trees: List[Tree]
        branchpoints: List[Branchpoint]
        comments: List[Comment]
        groups: List[Group]
    """

    parameters: NMLParameters
    trees: List[Tree]
    branchpoints: List[Branchpoint]
    comments: List[Comment]
    groups: List[Group]


def __parse_bounding_box(nml_parameters, prefix):
    boundingBox = None
    bboxName = prefix + "BoundingBox"
    if nml_parameters.find(bboxName) is not None:
        boundingBox = (
            int(nml_parameters.find(bboxName).get("topLeftX")),
            int(nml_parameters.find(bboxName).get("topLeftY")),
            int(nml_parameters.find(bboxName).get("topLeftZ")),
            int(nml_parameters.find(bboxName).get("width")),
            int(nml_parameters.find(bboxName).get("height")),
            int(nml_parameters.find(bboxName).get("depth")),
        )
    return boundingBox


def __parse_parameters(nml_parameters):
    offset = None
    if nml_parameters.find("offset") is not None:
        offset = (
            float(nml_parameters.find("offset").get("x")),
            float(nml_parameters.find("offset").get("y")),
            float(nml_parameters.find("offset").get("z")),
        )

    editRotation = None
    if nml_parameters.find("editRotation") is not None:
        editRotation = (
            float(nml_parameters.find("editRotation").get("xRot")),
            float(nml_parameters.find("editRotation").get("yRot")),
            float(nml_parameters.find("editRotation").get("zRot")),
        )

    editPosition = None
    if nml_parameters.find("editPosition") is not None:
        editPosition = (
            float(nml_parameters.find("editPosition").get("x")),
            float(nml_parameters.find("editPosition").get("y")),
            float(nml_parameters.find("editPosition").get("z")),
        )

    time = None
    if nml_parameters.find("time") is not None:
        time = int(nml_parameters.find("time").get("ms"))

    zoomLevel = None
    if nml_parameters.find("zoomLevel") is not None:
        zoomLevel = nml_parameters.find("zoomLevel").get("zoom")

    taskBoundingBox = __parse_bounding_box(nml_parameters, "task")
    userBoundingBox = __parse_bounding_box(nml_parameters, "user")

    return NMLParameters(
        name=nml_parameters.find("experiment").get("name"),
        scale=(
            float(nml_parameters.find("scale").get("x")),
            float(nml_parameters.find("scale").get("y")),
            float(nml_parameters.find("scale").get("z")),
        ),
        offset=offset,
        time=time,
        editPosition=editPosition,
        editRotation=editRotation,
        zoomLevel=zoomLevel,
        taskBoundingBox=taskBoundingBox,
        userBoundingBox=userBoundingBox,
    )


def __parse_node(nml_node):
    rotation = None
    if nml_node.get("rotX") is not None:
        rotation = (
            float(nml_node.get("rotX")),
            float(nml_node.get("rotY")),
            float(nml_node.get("rotZ")),
        )

    return Node(
        id=int(nml_node.get("id")),
        radius=float(nml_node.get("radius"))
        if nml_node.get("radius") is not None
        else None,
        position=(
            float(nml_node.get("x")),
            float(nml_node.get("y")),
            float(nml_node.get("z")),
        ),
        rotation=rotation,
        inVp=int(nml_node.get("inVp")) if nml_node.get("inVp") is not None else None,
        inMag=int(nml_node.get("inMag")) if nml_node.get("inMag") is not None else None,
        bitDepth=int(nml_node.get("bitDepth"))
        if nml_node.get("bitDepth") is not None
        else None,
        interpolation=bool(nml_node.get("interpolation"))
        if nml_node.get("interpolation") is not None
        else None,
        time=int(nml_node.get("time")) if nml_node.get("time") is not None else None,
    )


def __parse_edge(nml_edge):
    return Edge(source=int(nml_edge.get("source")), target=int(nml_edge.get("target")))


def __parse_tree(nml_tree):
    name = None
    if "comment" in nml_tree.attrib:
        name = nml_tree.get("comment")
    if "name" in nml_tree.attrib:
        name = nml_tree.get("name")

    color = None
    if "color.r" in nml_tree.attrib:
        color = (
            float(nml_tree.get("color.r")),
            float(nml_tree.get("color.g")),
            float(nml_tree.get("color.b")),
            float(nml_tree.get("color.a")),
        )
    if "colorr" in nml_tree.attrib:
        color = (
            float(nml_tree.get("colorr")),
            float(nml_tree.get("colorg")),
            float(nml_tree.get("colorb")),
            float(nml_tree.get("colora")),
        )
    try:
        groupId = int(nml_tree.get("groupId", default=-1))
    except ValueError:
        groupId = -1

    return Tree(
        nodes=[],
        edges=[],
        id=int(nml_tree.get("id")),
        name=name,
        groupId=groupId if groupId >= 0 else None,
        color=color,
    )


def __parse_branchpoint(nml_branchpoint):
    return Branchpoint(
        int(nml_branchpoint.get("id")),
        int(nml_branchpoint.get("time"))
        if nml_branchpoint.get("time") is not None
        else None,
    )


def __parse_comment(nml_comment):
    return Comment(
        int(nml_comment.get("node")), nml_comment.get("content", default=None)
    )


def __parse_group(nml_group):
    return Group(int(nml_group.get("id")), nml_group.get("name", default=None), [])


def parse_nml(file: BinaryIO) -> NML:
    """
    Reads a webKnossos NML skeleton file from disk, parses it and returns an NML Python object

    Attributes:
        file (BinaryIO): A Python file handle

    Return:
        NML: A webKnossos skeleton annotation as Python NML object

    Example:
        ```
        with open("input.nml", "rb") as f:
            nml = wknml.parse_nml(f, nml)
        ```
    """

    parameters = None
    trees = []
    branchpoints = []
    comments = []
    current_tree = None
    root_group = Group(-1, "", [])
    group_stack = [root_group]
    element_stack = []

    for event, elem in ET.iterparse(file, events=("start", "end")):
        if event == "start":
            element_stack.append(elem)
            if elem.tag == "thing":
                current_tree = __parse_tree(elem)
                trees.append(current_tree)
            elif elem.tag == "node":
                assert (
                    current_tree is not None
                ), "<node ...> tag needs to be child of a <thing ...> tag."
                current_tree.nodes.append(__parse_node(elem))
            elif elem.tag == "edge":
                assert (
                    current_tree is not None
                ), "<edge ...> tag needs to be child of a <thing ...> tag."
                current_tree.edges.append(__parse_edge(elem))
            elif elem.tag == "branchpoint":
                branchpoints.append(__parse_branchpoint(elem))
            elif elem.tag == "comment":
                comments.append(__parse_comment(elem))
            elif elem.tag == "group":
                group = __parse_group(elem)
                group_stack[-1].children.append(group)
                group_stack.append(group)
        elif event == "end":
            if elem.tag == "parameters":
                parameters = __parse_parameters(elem)
            elif elem.tag == "thing":
                current_tree = None
            elif elem.tag == "group":
                group_stack.pop()

            element_stack.pop()
            # Do not clear the elements of the parameters tag as we want to parse those all at once
            # when the closing parameters tag is parsed
            if len(element_stack) and element_stack[-1].tag != "parameters":
                # Discard the element to save memory
                elem.clear()

    return NML(
        parameters=parameters,
        trees=trees,
        branchpoints=branchpoints,
        comments=comments,
        groups=root_group.children,
    )


def __dump_bounding_box(xf, parameters, prefix):
    bboxName = prefix + "BoundingBox"
    parametersBox = getattr(parameters, bboxName)

    if parametersBox is not None:
        xf.tag(
            bboxName,
            {
                "topLeftX": str(parametersBox[0]),
                "topLeftY": str(parametersBox[1]),
                "topLeftZ": str(parametersBox[2]),
                "width": str(parametersBox[3]),
                "height": str(parametersBox[4]),
                "depth": str(parametersBox[5]),
            },
        )


def __dump_parameters(xf, parameters):
    xf.startTag("parameters")
    xf.tag("experiment", {"name": parameters.name})
    xf.tag(
        "scale",
        {
            "x": str(parameters.scale[0]),
            "y": str(parameters.scale[1]),
            "z": str(parameters.scale[2]),
        },
    )

    if parameters.offset is not None:
        xf.tag(
            "offset",
            {
                "x": str(parameters.offset[0]),
                "y": str(parameters.offset[1]),
                "z": str(parameters.offset[2]),
            },
        )

    if parameters.time is not None:
        xf.tag("time", {"ms": str(parameters.time)})
    if parameters.editPosition is not None:
        xf.tag(
            "editPosition",
            {
                "x": str(parameters.editPosition[0]),
                "y": str(parameters.editPosition[1]),
                "z": str(parameters.editPosition[2]),
            },
        )
    if parameters.editRotation is not None:
        xf.tag(
            "editRotation",
            {
                "xRot": str(parameters.editRotation[0]),
                "yRot": str(parameters.editRotation[1]),
                "zRot": str(parameters.editRotation[2]),
            },
        )
    if parameters.zoomLevel is not None:
        xf.tag("zoomLevel", {"zoom": str(parameters.zoomLevel)})

    __dump_bounding_box(xf, parameters, "task")
    __dump_bounding_box(xf, parameters, "user")

    xf.endTag()  # parameters


def __dump_node(xf, node):

    attributes = {
        "id": str(node.id),
        "x": str(node.position[0]),
        "y": str(node.position[1]),
        "z": str(node.position[2]),
    }

    if node.radius is not None:
        attributes["radius"] = str(node.radius)

    if node.rotation is not None:
        attributes["rotX"] = str(node.rotation[0])
        attributes["rotY"] = str(node.rotation[1])
        attributes["rotZ"] = str(node.rotation[2])

    if node.inVp is not None:
        attributes["inVp"] = str(node.inVp)

    if node.inMag is not None:
        attributes["inMag"] = str(node.inMag)

    if node.bitDepth is not None:
        attributes["bitDepth"] = str(node.bitDepth)

    if node.interpolation is not None:
        attributes["interpolation"] = str(node.interpolation)

    if node.time is not None:
        attributes["time"] = str(node.time)

    xf.tag("node", attributes)


def __dump_edge(xf, edge):
    xf.tag("edge", {"source": str(edge.source), "target": str(edge.target)})


def __dump_tree(xf, tree):
    attributes = {
        "id": str(tree.id),
        "color.r": str(tree.color[0]),
        "color.g": str(tree.color[1]),
        "color.b": str(tree.color[2]),
        "color.a": str(tree.color[3]),
        "name": tree.name,
    }

    if tree.groupId is not None:
        attributes["groupId"] = str(tree.groupId)

    xf.startTag("thing", attributes)
    xf.startTag("nodes")
    for n in tree.nodes:
        __dump_node(xf, n)
    xf.endTag()  # nodes
    xf.startTag("edges")
    for e in tree.edges:
        __dump_edge(xf, e)
    xf.endTag()  # edges
    xf.endTag()  # thing


def __dump_branchpoint(xf, branchpoint):
    if branchpoint.time is not None:
        xf.tag(
            "branchpoint", {"id": str(branchpoint.id), "time": str(branchpoint.time)}
        )
    else:
        xf.tag("branchpoint", {"id": str(branchpoint.id)})


def __dump_comment(xf, comment):
    if comment.content is not None:
        xf.tag("comment", {"node": str(comment.node), "content": comment.content})
    else:
        xf.tag("comment", {"node": str(comment.node)})


def __dump_group(xf, group):
    xf.startTag("group", {"id": str(group.id), "name": group.name})
    for g in group.children:
        __dump_group(xf, g)
    xf.endTag()  # group


def __dump_nml(xf, nml: NML):
    xf.startTag("things")
    __dump_parameters(xf, nml.parameters)
    for t in nml.trees:
        __dump_tree(xf, t)

    xf.startTag("branchpoints")
    for b in nml.branchpoints:
        __dump_branchpoint(xf, b)
    xf.endTag()  # branchpoints

    xf.startTag("comments")
    for c in nml.comments:
        __dump_comment(xf, c)
    xf.endTag()  # comments

    xf.startTag("groups")
    for g in nml.groups:
        __dump_group(xf, g)
    xf.endTag()  # groups
    xf.endTag()  # things


def write_nml(file: BinaryIO, nml: NML):
    """
    Writes an NML object to a file on disk.

        Arguments:
            file (BinaryIO): A Python file handle
            nml (NML): A NML object that should be persisted to disk

    Example:
        ```
        with open("out.nml", "wb") as f:
            wknml.write_nml(f, nml)
        ```
    """
    with XmlWriter(file) as xf:
        __dump_nml(xf, nml)
